#include "integer.hpp"

UInt::Data::Data(){
  nbref=1;
  length=1;
  digits=new Digit[1];
  digits[0]=0;
}

UInt::Data::Data(ulong l){
  nbref=1;
  length=l;
  digits=new Digit[l];
}

UInt::Data::Data(const list<Digit>& l){
  list<Digit>::const_iterator it;
  ulong i;
  nbref=1;
  length=l.size();
  digits=new Digit[length];
  i=0;
  for(it=l.begin();it!=l.end();++it){
    digits[i++]=*it;
  }
}

UInt::Data::~Data(){
  delete[] digits;
}

UInt::Data*
UInt::Data::get_own_copy(){
  Data *res;
  ulong i;
  if(nbref==1) return this;
  --nbref;
  res=new Data(length);
  for(i=0;i<length;++i) res->digits[i]=digits[i];
  return res;
}


UInt::UInt(){
  data=new Data;
}

UInt::UInt(const ulong& u){
  list<Digit> l;
  ulong t;
  Digit d;
  t=u;
  while(t>0){
    d=t&255;
    t>>=8;
    l.push_back(d);
  }
  data=new Data(l);
}

UInt::UInt(const UInt& u){
  data=u.data;
  ++data->nbref;
}

UInt::UInt(const list<Digit>& l){
  data=new Data(l);
}

UInt::~UInt(){
  if(--data->nbref==0) delete data;
}

const UInt&
UInt::operator=(const UInt& u){
  if(data==u.data) return *this;
  if(--data->nbref==0) delete data;
  data=u.data;
  ++data->nbref;
}

bool
UInt::operator==(const UInt &u) const{
  ulong i,l,lu;
  l=data->length;
  lu=u.data->length;
  if(l!=lu) return false;
  for(i=0;i<l;++i){
    if(data->digits[i]!=u.data->digits[i]) return false;
  }
  return true;
}

bool
UInt::operator<(const UInt &u) const{
  ulong i,k,l,lu;
  Digit d,du;
  l=data->length;
  lu=u.data->length;
  if(l<lu) return true;
  if(l>lu) return false;
  i=l-1;
  for(k=0;k<l;++k){
    d=data->digits[i];
    du=u.data->digits[i];
    if(d<du) return true;
    if(d>du) return false;
    --i;
  }
  return false;
}

bool
UInt::operator>(const UInt &u) const{
  ulong i,k,l,lu;
  Digit d,du;
  l=data->length;
  lu=u.data->length;
  if(l<lu) return false;
  if(l>lu) return true;
  i=l-1;
  for(k=0;k<l;++k){
    d=data->digits[i];
    du=u.data->digits[i];
    if(d<du) return false;
    if(d>du) return true;
    --i;
  }
  return false;
}

UInt 
UInt::operator+(const UInt& u) const{
  ulong i,l,lu,r,v,vu,w;
  Digit d;
  l=data->length;
  lu=u.data->length;
  if(l<lu) return u.operator+(*this);
  //We have l>=lu
  list<Digit> lres;
  r=0;
  for(i=0;i<lu;++i){
    v=data->digits[i];
    vu=u.data->digits[i];
    w=v+vu+r;
    if(w>256){
      w-=256;
      r=1;
    }
    else r=0;
    lres.push_back(w);
  }
  for(i=lu;i<l;++i){
    v=data->digits[i];
    w=v+r;
    if(w>256){
      w-=256;
      r=1;
    }
    else r=0;
    lres.push_back(w);
  }
  if(r!=0) lres.push_back(r);
  return UInt(lres);
}

UInt
UInt::operator-(const UInt& u) const{
  ulong i,l,lu,r,v,vu,w;
  l=data->length;
  lu=data->length;
  if(lu>l) throw("UInt::operator-(Uint) : the result is negative");
  list<Digit> lres;
  r=0;
  for(i=0;i<lu;++i){
    v=data->digits[i];
    vu=u.data->digits[i]+r;
    if(v>=vu){
      w=v-vu;
      r=0;
    }
    else{
      w=v+(256-vu);
      r=1;
    }
    lres.push_back(w);
  }
  for(i=lu;i<l;++i){
    v=data->digits[i];
    if(v>=r){
      w=v-r;
      r=0;
    }
    else{
      w=v+(256-r);
      r=1;
    }
    lres.push_back(w);
  }
  if(r!=0) throw("UInt::operator-(Uint) : the result is negative");
  //Remove the 0's
  list<Digit>::iterator it=lres.end();
  --it;
  while(*it==0 and it!=lres.begin()) --it;
  ++it;
  lres.erase(it,lres.end());
  return UInt(lres);
}

UInt 
UInt::operator*(const Digit& d) const{
  if(d==0) return UInt();
  ulong i,l,r,v,w;
  list<Digit> lres;
  l=data->length;
  r=0;
  for(i=0;i<l;++i){
    v=data->digits[i];
    w=v*d+r;
    r=w/256;
    w%=256;
    lres.push_back(w);
  }
  if(r!=0) lres.push_back(r);
  return UInt(lres);
}

UInt 
UInt::operator*(const UInt& u) const{
  ulong i,j,k,l,lu,r,v,w;
  UInt temp,res;
  Digit d;
  list<Digit> lres;
  l=data->length;
  lu=u.data->length;
  for(i=0;i<lu;++i){
    d=u.data->digits[i];
    if(d!=0){
      for(k=0;k<i;++k) lres.push_back(0);
      r=0;
      for(j=0;j<l;++j){
	v=data->digits[j];
	w=v*d+r;
	r=w/256;
	w%=256;
	lres.push_back(w);
      }
      if(r!=0) lres.push_back(r);
      temp=lres;
      lres.clear();
      res=res+temp;
    }
  }
  return res;
}

UInt
UInt::mul_and_shift(const Digit& d,const ulong& s) const{
  ulong i,l,r,v,w;
  list<Digit> lres;
  for(i=0;i<s;++i) lres.push_back(0);
  l=data->length;
  r=0;
  for(i=0;i<l;++i){
    v=data->digits[i];
    w=v*d+r;
    r=w/256;
    w%=256;
    lres.push_back(w);
  }
  if(r!=0) lres.push_back(r);
  return UInt(lres);
  
}

pair<UInt,UInt>
UInt::division(const UInt& D) const{
  if(*this<D) return pair<UInt,UInt>(0,D);
  ulong cN,cD,i,lN,lD,lT,ldelta,pldelta,q;
  UInt N=*this;
  UInt T;
  list<Digit> lQ;
  lD=D.data->length;
  lN=N.data->length;
  cD=D.data->digits[lD-1];
  cN=N.data->digits[lN-1];
  ldelta=lN-lD;
  pldelta=ldelta;
  while(N>D){
    lN=N.data->length;
    cN=N.data->digits[lN-1];
    ldelta=lN-lD;
    if(cD>cN){
      cN=cN*256+N.data->digits[lN-2];
      --ldelta;
    }
    q=cN/cD;
    T=D.mul_and_shift(q,ldelta);
    while(T>N){
      T=T-D;
      --q;
    }
    for(i=pldelta;i<ldelta;++i) lQ.push_back(0);
    lQ.push_back(q);
    N=N-T;
    pldelta=ldelta;
  }
  return pair<UInt,UInt>(lQ,N);
}

ostream&
operator<<(ostream& os,const UInt& u){
  ulong i,l;
  l=u.data->length;
  os<<'[';
  os<<(ulong)u.data->digits[l-1];
  for(i=1;i<u.data->length;++i){
    os<<','<<(ulong)u.data->digits[l-i-1];
  }
  os<<']';
  return os;
}

  
